from .modelout_functions import AbstractModelOutput, TASK_TO_MODELOUT
from .projectors import ProjectionType, AbstractProjector, CudaProjector, BasicProjector
from .gradient_computers import FunctionalGradientComputer,\
                                AbstractGradientComputer
from .score_computers import AbstractScoreComputer, BasicScoreComputer
from .savers import AbstractSaver, MmapSaver, ModelIDException
from .utils import get_num_params

from typing import Iterable, Optional, Union
from pathlib import Path
from tqdm import tqdm
from torch import Tensor

import numpy as np
import torch
ch = torch


class TRAKer():
    """ The main front-facing class for TRAK. See the `README
    <https://github.com/MadryLab/trak>`_ and `docs
    <https://trak.readthedocs.io/en/latest/>`_ for example usage.

    """
    def __init__(self,
                 model: torch.nn.Module,
                 task: Union[AbstractModelOutput, str],
                 train_set_size: int,
                 save_dir: str = './trak_results',
                 load_from_save_dir: bool = True,
                 device: Union[str, torch.device] = 'cuda',
                 gradient_computer: AbstractGradientComputer = FunctionalGradientComputer,
                 projector: Optional[AbstractProjector] = None,
                 saver: Optional[AbstractSaver] = None,
                 score_computer: Optional[AbstractScoreComputer] = None,
                 proj_dim: int = 2048,
                 ) -> None:
        """

        Args:
            model (torch.nn.Module):
                model to use for TRAK
            task (Union[AbstractModelOutput, str]):
                Type of model that TRAK will be ran on. Accepts either one of
                the following strings:
                - :code:`image_classification`
                - :code:`text_classification`
                - :code:`clip`
                or an implementation of :class:`.AbstractModelOutput`.
            train_set_size (int):
                Size of the train set that TRAK is featurizing
            save_dir (str, optional):
                Directory to save final TRAK scores, intermediate results, and
                metadata. Defaults to :code:'./trak_results'.
            load_from_save_dir (bool, optional):
                If True, the :class`.TRAKer` instance will attempt to load
                existing metadata from save_dir. May lead to I/O issues if
                multiple TRAKer instances ran in parallel have this flag set to
                True. See the SLURM tutorial for more details.
            device (Union[str, torch.device], optional):
                torch device on which to do computations. Defaults to 'cuda'.
            gradient_computer (AbstractGradientComputer, optional):
                Class to use to get per-example gradients. See
                :class:`.AbstractGradientComputer` for more details. Defaults to
                :class:`.FunctionalGradientComputer`.
            projector (Optional[AbstractProjector], optional):
                Either set :code:`proj_dim` and a :class:`.CudaProjector`
                Rademacher projector will be used or give a custom subclass of
                :class:`.AbstractProjector` class and leave :code:`proj_dim` as
                None. Defaults to None.
            saver (Optional[AbstractSaver], optional):
                Class to use for saving intermediate results and final TRAK
                scores to RAM/disk. If None, the :class:`.MmapSaver` will
                be used. Defaults to None.
            score_computer (Optional[AbstractScoreComputer], optional):
                Class to use for computing the final TRAK scores. If None, the
                :class:`.BasicScoreComputer` will be used. Defaults to None.
            proj_dim (int, optional):
                Dimension of the projected TRAK features. See Section 4.3 of
                `our paper <https://arxiv.org/abs/2303.14186>`_ for more
                details. Defaults to 2048.

        """

        self.model = model
        self.task = task
        self.train_set_size = train_set_size
        self.device = device

        self.num_params = get_num_params(self.model)
        self.init_projector(projector, proj_dim)  # inits self.projector

        self.save_dir = Path(save_dir).resolve()
        self.load_from_save_dir = load_from_save_dir

        if type(self.task) is str:
            self.task = TASK_TO_MODELOUT[(self.task, gradient_computer.is_functional)]

        self.gradient_computer = gradient_computer(model=self.model,
                                                   modelout_fn=self.task,
                                                   grad_dim=self.num_params)

        if score_computer is None:
            score_computer = BasicScoreComputer
        self.score_computer = score_computer(device=self.device)

        metadata = {
            'JL dimension': self.projector.proj_dim,
            'JL matrix type': self.projector.proj_type,
        }

        if saver is None:
            saver = MmapSaver
        self.saver = saver(save_dir=self.save_dir,
                           metadata=metadata,
                           train_set_size=self.train_set_size,
                           proj_dim=self.proj_dim,
                           load_from_save_dir=self.load_from_save_dir)

    def init_projector(self, projector, proj_dim) -> None:
        """ Initialize the projector for a traker class

        Args:
            projector (AbstractProjector):
                JL projector

        """

        self.projector = projector
        if projector is not None:
            self.proj_dim = self.projector.proj_dim
        else:
            self.proj_dim = proj_dim
            try:
                import fast_jl
                test_gradient = ch.ones(1, self.num_params).cuda()
                num_sms = ch.cuda.get_device_properties('cuda').multi_processor_count
                fast_jl.project_rademacher_8(test_gradient, self.proj_dim, 0, num_sms)
                projector = CudaProjector

            except (ImportError, RuntimeError, AttributeError) as e:
                print(f'Could not use CudaProjector.\nReason: {str(e)}')
                print('Defaulting to BasicProjector.')
                projector = BasicProjector

            self.projector = projector(grad_dim=self.num_params,
                                       proj_dim=self.proj_dim,
                                       seed=0,
                                       proj_type=ProjectionType.rademacher,
                                       device=self.device)

    def load_checkpoint(self,
                        checkpoint: Iterable[Tensor],
                        model_id: int,
                        _allow_featurizing_already_registered=None) -> None:
        """ Loads state dictionary for the given checkpoint; initializes arrays
        to store TRAK features for that checkpoint, tied to the model ID.

        Args:
            checkpoint (Iterable[Tensor]):
                state_dict to load
            model_id (int):
                a unique ID for a checkpoint
            _allow_featurizing_already_registered (bool, optional):
                Only use if you want to override the default behaviour that
                :code:`featurize` is forbidden on already registered model IDs.
                Defaults to None.

        """
        if self.saver.model_ids.get(model_id) is None:
            self.saver.register_model_id(model_id,
                                         _allow_featurizing_already_registered)
        else:
            self.saver.load_store(model_id)

        self.model.load_state_dict(checkpoint)
        self.model.eval()

        self.gradient_computer.load_model_params(self.model)

        self._last_ind = 0
        self._num_featurized = 0

    def featurize(self,
                  batch: Iterable[Tensor],
                  inds: Optional[Iterable[int]] = None,
                  num_samples: Optional[int] = None
                  ) -> None:
        """ Creates TRAK features for the given batch by computing the gradient
        of the model output function and projecting it. In the notation of the
        paper, for an input pair :math:`z=(x,y)`, model parameters
        :math:`\\theta`, and JL projection matrix :math:`P`, this method
        computes :math:`P^\\top \\nabla_\\theta f(z_i, \\theta)`.
        Additionally, this method computes the gradient of the out-to-loss
        function (in the notation of the paper, the :math:`Q` term in Section
        3.4).

        Either :code:`inds` or :code:`num_samples` must be specified. Using
        :code:`num_samples` will write sequentially into the internal store of
        the :func:`TRAKer`.

        Args:
            batch (Iterable[Tensor]):
                input batch
            inds (Optional[Iterable[int]], optional):
                Indices of the batch samples in the train set. Defaults to None.
            num_samples (Optional[int], optional):
                Number of samples in the batch. Defaults to None.

        """
        assert (inds is None) or (num_samples is None),\
            "Exactly one of num_samples and inds should be specified"
        assert (inds is not None) or (num_samples is not None),\
            "Exactly one of num_samples and inds should be specified"

        if num_samples is not None:
            inds = np.arange(self._last_ind, self._last_ind + num_samples)
            self._last_ind += num_samples
        else:
            num_samples = inds.reshape(-1).shape[0]
        self._num_featurized += num_samples

        grads = self.gradient_computer.compute_per_sample_grad(batch=batch,
                                                               batch_size=num_samples)
        grads = self.projector.project(grads, model_id=self.saver.current_model_id)
        self.saver.current_grads[inds] = grads.cpu().clone().detach()

        loss_grads = self.gradient_computer.compute_loss_grad(batch)
        self.saver.current_out_to_loss[inds] = loss_grads.cpu().clone().detach()

        if self._num_featurized == self.train_set_size:
            # TODO: currently this is breaking abstraction -- either define dict
            # items in abstract __init__, or don't access them here
            self.saver.model_ids[self.saver.current_model_id]['featurized'] = 1
            self.saver.serialize_model_id_metadata(self.saver.current_model_id)

    def finalize_features(self,
                          model_ids: Iterable[int] = None,
                          del_grads: bool = False) -> None:
        """ For a set of checkpoints :math:`C` (specified by model IDs), and
        gradients :math:`\\{ \\Phi_c \\}_{c\\in C}`, this method computes
        :math:`\\Phi_c (\\Phi_c^\\top\\Phi_c)^{-1}` for all :math:`c\\in C`
        and stores the results in the internal store of the :func:`TRAKer`
        class.

        Args:
            model_ids (Iterable[int], optional): A list of model IDs for which
                features should be finalized. If None, features are finalized
                for all model IDs in the :code:`save_dir` of the :class:`.TRAKer`
                class. Defaults to None.

        """
        if model_ids is None:
            model_ids = list(self.saver.model_ids.keys())

        self._last_ind = 0

        for model_id in tqdm(model_ids, desc='Finalizing features for all model IDs..'):
            if self.saver.model_ids.get(model_id) is None:
                raise ModelIDException(f'Model ID {model_id} not registered, not ready for finalizing.')
            elif self.saver.model_ids[model_id]['finalized'] == 1:
                print(f'Model ID {model_id} already finalized, skipping .finalize_features for it.')
                continue

            self.saver.load_store(model_id)

            g = ch.as_tensor(self.saver.current_grads)
            xtx = self.score_computer.get_xtx(g)

            self.saver.current_features[:] = self.score_computer.get_x_xtx_inv(g, xtx).cpu()
            if del_grads:
                self.saver.del_grads(model_id)

            self.saver.model_ids[self.saver.current_model_id]['finalized'] = 1
            self.saver.serialize_model_id_metadata(self.saver.current_model_id)

    def start_scoring_checkpoint(self,
                                 checkpoint: Iterable[Tensor],
                                 model_id: int,
                                 num_targets: int,
                                 ) -> None:
        """ This method prepares the internal store of the :class:`.TRAKer` class
        to start computing scores for a set of targets.

        Args:
            checkpoint (Iterable[Tensor]):
                model checkpoint (state dict)
            model_id (int):
                a unique ID for a checkpoint
            num_targets (int):
                number of targets to score

        """
        self.saver.load_target_store(model_id, num_targets, mode='w+')

        self.model.load_state_dict(checkpoint)
        self.model.eval()
        self.gradient_computer.load_model_params(self.model)

        self._last_ind_target = 0

    def score(self,
              batch: Iterable[Tensor],
              inds: Optional[Iterable[int]] = None,
              num_samples: Optional[int] = None,
              ) -> None:
        """ This method computes the (intermediate per-checkpoint) TRAK scores
        for a batch of targets and stores them in the internal store of the
        :class:`.TRAKer` class.

        Either :code:`inds` or :code:`num_samples` must be specified. Using
        :code:`num_samples` will write sequentially into the internal store of
        the :class:`.TRAKer`.

        Args:
            batch (Iterable[Tensor]):
                input batch
            inds (Optional[Iterable[int]], optional):
                Indices of the batch samples in the train set. Defaults to None.
            num_samples (Optional[int], optional):
                Number of samples in the batch. Defaults to None.

        """
        assert (inds is None) or (num_samples is None),\
            "Exactly one of num_samples and inds should be specified"
        assert (inds is not None) or (num_samples is not None),\
            "Exactly one of num_samples and inds should be specified"

        # TODO: currently this is breaking abstraction -- either define dict
        # items in abstract __init__, or don't access them here
        if self.saver.model_ids[self.saver.current_model_id]['finalized'] == 0:
            print(f'Model ID {self.saver.current_model_id} not finalized, cannot score')
            return None

        if num_samples is not None:
            inds = np.arange(self._last_ind_target, self._last_ind_target + num_samples)
            self._last_ind_target += num_samples
        else:
            num_samples = inds.reshape(-1).shape[0]

        grads = self.gradient_computer.compute_per_sample_grad(batch=batch,
                                                               batch_size=num_samples)

        grads = self.projector.project(grads, model_id=self.saver.current_model_id)

        self.saver.current_target_grads[inds] = grads.cpu().clone().detach()

    def finalize_scores(self,
                        model_ids: Iterable[int] = None,
                        del_grads: bool = False,
                        allow_skip: bool = False,
                        exp_name: str = None) -> Tensor:
        """ This method computes the final TRAK scores for the given targets,
        train samples, and model checkpoints (specified by model IDs).

        Args:
            model_ids (Iterable[int], optional):
                A list of model IDs for which
                scores should be finalized. If None, scores are computed
                for all model IDs in the :code:`save_dir` of the :class:`.TRAKer`
                class. Defaults to None.
            del_grads (bool, optional):
                If True, the target gradients (intermediate results) are deleted
                from the internal store of the :class:`.TRAKer` class.  Defaults
                to False.
            allow_skip (bool, optional):
                If True, raises only a warning, instead of an error, when target
                gradients are not computed for a given model ID. Defaults to
                False.
            exp_name (str, optional):
                Used to name the scores :code:`.npy` array produced by this
                method in the :code:`save_dir` of the :class:`.TRAKer` class. If
                None, a random uuid is generated.  Defaults to None.

        Returns:
            Tensor: TRAK scores

        """
        # reset counter for inds used for scoring
        self._last_ind_target = 0

        if model_ids is None:
            model_ids = self.saver.model_ids

        _completed = [False] * len(model_ids)
        _scores = ch.zeros(self.train_set_size,
                           self.saver.num_targets,
                           device=self.device)
        _avg_out_to_losses = ch.zeros(self.saver.train_set_size, 1, device=self.device)

        for j, model_id in enumerate(tqdm(model_ids, desc='Finalizing scores for all model IDs..')):
            self.saver.load_store(model_id)
            try:
                self.saver.load_target_store(model_id, self.saver.num_targets)
            except OSError as e:
                if allow_skip:
                    print(f'Could not read target gradients for model ID {model_id}. Skipping.')
                    continue
                else:
                    raise e

            # TODO: currently this is breaking abstraction -- either define dict
            # items in abstract __init__, or don't access them here
            if self.saver.model_ids[self.saver.current_model_id]['finalized'] == 0:
                print(f'Model ID {self.saver.current_model_id} not finalized, cannot score')
                continue

            g = ch.as_tensor(self.saver.current_features, device=self.device)
            g_target = ch.as_tensor(self.saver.current_target_grads, device=self.device)

            _scores += self.score_computer.get_scores(g, g_target)
            _avg_out_to_losses += ch.as_tensor(self.saver.current_out_to_loss, device=self.device)
            _completed[j] = True

            if del_grads:
                self.saver.del_grads(model_id, target=True)
            else:
                self.saver.clear_target_grad_count(model_id)

        _num_models_used = float(sum(_completed))
        self.scores = (_scores / _num_models_used) * (_avg_out_to_losses / _num_models_used)
        self.saver.save_scores(self.scores.cpu().numpy(), exp_name)

        return self.scores



import pytest
from tqdm import tqdm
from pathlib import Path
import torch as ch

from trak import TRAKer
from .utils import construct_rn9, get_dataloader, eval_correlations
from .utils import download_cifar_checkpoints, download_cifar_betons


@pytest.mark.cuda
def test_featurize_and_score_in_parallel(tmp_path):
    device = 'cuda:0'
    batch_size = 100

    model = construct_rn9().to(memory_format=ch.channels_last).to(device)
    model = model.eval()

    BETONS_PATH = Path(tmp_path).joinpath('cifar_betons')
    BETONS = download_cifar_betons(BETONS_PATH)

    loader_train = get_dataloader(BETONS, batch_size=batch_size, split='train')
    loader_val = get_dataloader(BETONS, batch_size=batch_size, split='val')

    CKPT_PATH = Path(tmp_path).joinpath('cifar_ckpts')
    ckpt_files = download_cifar_checkpoints(CKPT_PATH)
    ckpts = [ch.load(ckpt, map_location='cpu') for ckpt in ckpt_files]

    # this should be essentially equivalent to running each
    # TRAKer in a separate script
    for model_id, ckpt in enumerate(ckpts):
        traker = TRAKer(model=model,
                        task='image_classification',
                        train_set_size=10_000,
                        save_dir=tmp_path,
                        device=device)
        traker.load_checkpoint(checkpoint=ckpt, model_id=model_id)
        for batch in tqdm(loader_train, desc='Computing TRAK embeddings...'):
            traker.featurize(batch=batch, num_samples=len(batch[0]))
        traker.finalize_features()

    for model_id, ckpt in enumerate(ckpts):
        traker = TRAKer(model=model,
                        task='image_classification',
                        train_set_size=10_000,
                        save_dir=tmp_path,
                        device=device)

        traker.start_scoring_checkpoint(ckpt, model_id, num_targets=2_000)
        for batch in tqdm(loader_val, desc='Scoring...'):
            traker.score(batch=batch, num_samples=len(batch[0]))

    scores = traker.finalize_scores().cpu()

    avg_corr = eval_correlations(infls=scores, tmp_path=tmp_path)
    assert avg_corr > 0.062, 'correlation with 3 CIFAR-2 models should be >= 0.062'


@pytest.mark.cuda
def test_score_multiple(tmp_path):
    device = 'cuda:0'
    batch_size = 100

    model = construct_rn9().to(memory_format=ch.channels_last).to(device)
    model = model.eval()

    BETONS_PATH = Path(tmp_path).joinpath('cifar_betons')
    BETONS = download_cifar_betons(BETONS_PATH)

    loader_train = get_dataloader(BETONS, batch_size=batch_size, split='train')
    loader_val = get_dataloader(BETONS, batch_size=batch_size, split='val')

    CKPT_PATH = Path(tmp_path).joinpath('cifar_ckpts')
    ckpt_files = download_cifar_checkpoints(CKPT_PATH)
    ckpts = [ch.load(ckpt, map_location='cpu') for ckpt in ckpt_files]

    traker = TRAKer(model=model,
                    task='image_classification',
                    train_set_size=10_000,
                    save_dir=tmp_path,
                    device=device)

    for model_id, ckpt in enumerate(ckpts):
        traker.load_checkpoint(checkpoint=ckpt, model_id=model_id)
        for batch in tqdm(loader_train, desc='Computing TRAK embeddings...'):
            traker.featurize(batch=batch, num_samples=len(batch[0]))
    traker.finalize_features()

    scoring_runs = range(3)
    for _ in scoring_runs:
        for model_id, ckpt in enumerate(ckpts):
            traker = TRAKer(model=model,
                            task='image_classification',
                            train_set_size=10_000,
                            save_dir=tmp_path,
                            device=device)

            traker.start_scoring_checkpoint(ckpt, model_id, num_targets=2_000)
            for batch in tqdm(loader_val, desc='Scoring...'):
                traker.score(batch=batch, num_samples=len(batch[0]))

        scores = traker.finalize_scores().cpu()

        avg_corr = eval_correlations(infls=scores, tmp_path=tmp_path)
        assert avg_corr > 0.062, 'correlation with 3 CIFAR-2 models should be >= 0.062'



import pytest
from tqdm import tqdm
from torchvision import datasets
import open_clip

from trak import TRAKer
# from trak.gradient_computers import IterativeGradientComputer


@pytest.mark.cuda
def test_mscoco(tmp_path, device='cuda:0'):
    model, _, preprocess = open_clip.create_model_and_transforms('RN50')
    model = model.to(device)
    model.eval()

    tokenizer = open_clip.get_tokenizer('RN50')

    ds_train = datasets.CocoCaptions(root='/path/to/coco_csv/train2014',
                                     annFile='/path/to/coco_csv/coco_train_karpathy.json'
                                     )

    traker = TRAKer(model=model,
                    task='clip',
                    save_dir=tmp_path,
                    train_set_size=len(ds_train),
                    device=device,
                    proj_dim=512,
                    )

    traker.modelout_fn.get_embeddings(model, ds_train, batch_size=1, size=600,
                                      preprocess_fn_img=lambda x: preprocess(x).to(device).unsqueeze(0),
                                      preprocess_fn_txt=lambda x: tokenizer(x[0]).to(device))

    traker.load_checkpoint(model.state_dict(), model_id=0)
    for bind, (img, captions) in enumerate(tqdm(ds_train)):
        x = preprocess(img).to(device).unsqueeze(0)
        # selecting (wlog) the first out of 5 captions
        y = tokenizer(captions[0]).to(device)

        traker.featurize(batch=(x, y), num_samples=x.shape[0])
        if bind == 2:
            break

    traker.finalize_features()



from trak import TRAKer
from torchvision.models import resnet18
import pytest
import torch as ch
from trak.projectors import BasicProjector
from trak.modelout_functions import ImageClassificationModelOutput


@pytest.fixture
def cpu_proj():
    projector = BasicProjector(grad_dim=11689512,
                               proj_dim=20,
                               seed=0,
                               proj_type='rademacher',
                               device='cpu')
    return projector


def test_class_init(tmp_path, cpu_proj):
    model = resnet18()
    TRAKer(model=model,
           task='image_classification',
           save_dir=tmp_path,
           projector=cpu_proj,
           train_set_size=20,
           device='cuda:0')


def test_load_ckpt(tmp_path, cpu_proj):
    model = resnet18()
    traker = TRAKer(model=model,
                    task='image_classification',
                    save_dir=tmp_path,
                    projector=cpu_proj,
                    train_set_size=20,
                    device='cuda:0')
    ckpt = model.state_dict()
    traker.load_checkpoint(ckpt, model_id=0)


def test_load_ckpt_repeat(tmp_path, cpu_proj):
    model = resnet18()
    traker = TRAKer(model=model,
                    task='image_classification',
                    save_dir=tmp_path,
                    projector=cpu_proj,
                    train_set_size=20,
                    device='cuda:0')
    ckpt = model.state_dict()
    traker.load_checkpoint(ckpt, model_id=0)
    traker.load_checkpoint(ckpt, model_id=1)


@pytest.mark.cuda
def test_featurize(tmp_path):
    model = resnet18().cuda().eval()
    N = 5
    batch = ch.randn(N, 3, 32, 32).cuda(), ch.randint(low=0, high=10, size=(N,)).cuda()
    traker = TRAKer(model=model,
                    task='image_classification',
                    save_dir=tmp_path,
                    train_set_size=20,
                    device='cuda:0')
    ckpt = model.state_dict()
    traker.load_checkpoint(ckpt, model_id=0)
    traker.featurize(batch, num_samples=N)


@pytest.mark.cuda
def test_finalize_features(tmp_path):
    model = resnet18().cuda().eval()
    N = 5
    batch = ch.randn(N, 3, 32, 32).cuda(), ch.randint(low=0, high=10, size=(N,)).cuda()
    traker = TRAKer(model=model,
                    task='image_classification',
                    save_dir=tmp_path,
                    train_set_size=N,
                    device='cuda:0')
    ckpt = model.state_dict()
    traker.load_checkpoint(ckpt, model_id=0)
    traker.featurize(batch, num_samples=N)
    traker.finalize_features()


@pytest.mark.cuda
def test_score(tmp_path):
    model = resnet18().cuda().eval()
    N = 5
    batch = ch.randn(N, 3, 32, 32).cuda(), ch.randint(low=0, high=10, size=(N,)).cuda()
    traker = TRAKer(model=model,
                    task='image_classification',
                    save_dir=tmp_path,
                    train_set_size=N,
                    device='cuda:0')
    ckpt = model.state_dict()
    traker.load_checkpoint(ckpt, model_id=0)
    traker.featurize(batch, num_samples=N)
    traker.finalize_features()
    traker.start_scoring_checkpoint(ckpt, 0, num_targets=N)
    traker.score(batch, num_samples=N)


@pytest.mark.cuda
def test_score_finalize(tmp_path):
    model = resnet18().cuda().eval()
    N = 5
    batch = ch.randn(N, 3, 32, 32).cuda(), ch.randint(low=0, high=10, size=(N,)).cuda()
    traker = TRAKer(model=model,
                    task='image_classification',
                    save_dir=tmp_path,
                    train_set_size=N,
                    device='cuda:0')
    ckpt = model.state_dict()
    traker.load_checkpoint(ckpt, model_id=0)
    traker.featurize(batch, num_samples=N)
    traker.finalize_features()

    traker.start_scoring_checkpoint(ckpt, 0, num_targets=N)
    traker.score(batch, num_samples=N)
    traker.finalize_scores()


def test_custom_model_output(tmp_path, cpu_proj):
    model = resnet18()
    TRAKer(model=model,
           task=ImageClassificationModelOutput,
           save_dir=tmp_path,
           projector=cpu_proj,
           train_set_size=20,
           device='cuda:0')



from tqdm import tqdm
from pathlib import Path
from itertools import product
from scipy.stats import spearmanr
import pytest
import torch
import torchvision
import numpy as np

from trak import TRAKer
from trak.projectors import BasicProjector

from .utils import construct_rn9


ch = torch


def get_dataloader(batch_size=256, num_workers=8, split='train', shuffle=False, augment=True):
    if augment:
        transforms = torchvision.transforms.Compose(
                        [torchvision.transforms.RandomHorizontalFlip(),
                         torchvision.transforms.RandomAffine(0),
                         torchvision.transforms.ToTensor(),
                         torchvision.transforms.Normalize((0.4914, 0.4822, 0.4465),
                                                          (0.2023, 0.1994, 0.201))])
    else:
        transforms = torchvision.transforms.Compose([
                         torchvision.transforms.ToTensor(),
                         torchvision.transforms.Normalize((0.4914, 0.4822, 0.4465),
                                                          (0.2023, 0.1994, 0.201))])

    is_train = (split == 'train')
    dataset = torchvision.datasets.CIFAR10(root='/tmp/cifar/',
                                           download=True,
                                           train=is_train,
                                           transform=transforms)

    loader = torch.utils.data.DataLoader(dataset=dataset,
                                         shuffle=shuffle,
                                         batch_size=batch_size,
                                         num_workers=num_workers)

    return loader


def eval_correlations(infls, tmp_path):

    masks_path = Path('/mnt/cfs/projects/better_tracin/eval_margins/CIFAR10/50pct/mask.npy')
    # num masks, num train samples
    masks = ch.as_tensor(np.load(masks_path, mmap_mode='r')).float()

    margins_path = Path('/mnt/cfs/projects/better_tracin/eval_margins/CIFAR10/50pct/val_margins.npy')
    # num , num val samples
    margins = ch.as_tensor(np.load(margins_path, mmap_mode='r'))

    val_inds = np.arange(2000)
    preds = masks @ infls
    rs = []
    ps = []
    for ind, j in tqdm(enumerate(val_inds)):
        r, p = spearmanr(preds[:, ind], margins[:, j])
        rs.append(r)
        ps.append(p)
    rs, ps = np.array(rs), np.array(ps)
    print(f'Correlation: {rs.mean()} (avg p value {ps.mean()})')
    return rs.mean()


def get_projector(use_cuda_projector, dtype):
    if use_cuda_projector:
        return None
    return BasicProjector(grad_dim=2274880, proj_dim=4096,
                          seed=0, proj_type='normal', block_size=400,
                          dtype=dtype, device='cuda:0')


PARAM = list(product([True],  # serialize
                     [True],  # basic / cuda projector
                     [ch.float16],  # projection dtype
                     [100],  # batch size
                     ))


@pytest.mark.parametrize("serialize, use_cuda_projector, dtype, batch_size", PARAM)
@pytest.mark.cuda
def test_cifar_acc(serialize, use_cuda_projector, dtype, batch_size, tmp_path):
    device = 'cuda:0'
    projector = get_projector(use_cuda_projector, dtype)
    model = construct_rn9(10).to(memory_format=ch.channels_last).to(device)
    model = model.eval()

    loader_train = get_dataloader(batch_size=batch_size, split='train', augment=False)
    loader_val = get_dataloader(batch_size=batch_size, split='val', augment=False)

    #  = Path(tmp_path).joinpath('cifar_ckpts')
    # ckpt_files = download_cifar_checkpoints(CKPT_PATH)

    # ckpt_files = sorted(list(Path('/mnt/xfs/home/krisgrg/projects/trak/examples/checkpoints_1').rglob('*.pt')))
    ckpt_files = sorted(list(Path('/mnt/xfs/home/krisgrg/projects/trak/examples/checkpoints').rglob('*.pt')))

    ckpts = [ch.load(ckpt, map_location='cpu') for ckpt in ckpt_files]

    traker = TRAKer(model=model,
                    task='image_classification',
                    projector=projector,
                    proj_dim=4096,
                    train_set_size=50_000,
                    save_dir=tmp_path,
                    device=device)

    print('GRAD DIM:', traker.projector.grad_dim)
    # print('jl matrix:', traker.projector.proj_matrix)

    for model_id, ckpt in enumerate(ckpts):
        traker.load_checkpoint(checkpoint=ckpt, model_id=model_id)
        for batch in tqdm(loader_train, desc='Computing TRAK embeddings...'):
            batch = [x.cuda() for x in batch]
            traker.featurize(batch=batch, num_samples=len(batch[0]))

    traker.finalize_features()

    # print('jl matrix:', traker.projector.proj_matrix)

    if serialize:
        del traker
        traker = TRAKer(model=model,
                        task='image_classification',
                        projector=projector,
                        proj_dim=4096,
                        train_set_size=50_000,
                        save_dir=tmp_path,
                        device=device)

    for model_id, ckpt in enumerate(ckpts):
        traker.start_scoring_checkpoint(ckpt, model_id, num_targets=10_000)
        for batch in tqdm(loader_val, desc='Scoring...'):
            batch = [x.cuda() for x in batch]
            traker.score(batch=batch, num_samples=len(batch[0]))

    scores = traker.finalize_scores().cpu()

    avg_corr = eval_correlations(infls=scores, tmp_path=tmp_path)
    assert avg_corr > 0.062, 'correlation with 3 CIFAR-2 models should be >= 0.062'

    # self.scores = (_scores / _num_models_used)
    # print('NO OUT-TO-LOSS GRADS')


import pytest
import math
from itertools import product
import numpy as np
import torch as ch
from torch import testing

from trak.projectors import CudaProjector, ProjectionType
BasicProjector = CudaProjector

PARAM = list(product([0, 1, 10**8],  # seed
                     [ProjectionType.normal, ProjectionType.rademacher],  # proj type
                     [ch.float16, ch.float32],  # dtype
                     [
                         (1, 25),
                         (8, 10_000),
                         (16, 10_002),
                         (9, 10_002),
                         (16, 10_001),
                         (45, 1049),
                         (1, int(1e9)),
                     ],  # input shape
                     [4096, 1024],  # proj dim
                     ))


@pytest.mark.parametrize("seed, proj_type, dtype, input_shape, proj_dim", PARAM)
@pytest.mark.cuda
def test_seed_consistency(seed,
                          proj_type,
                          dtype,
                          proj_dim,
                          input_shape,
                          ):
    """
    Check that re-running the same projection with the same seed
    leads to the same result.
    """

    g = testing.make_tensor(*input_shape, device='cuda:0', dtype=dtype)
    proj = BasicProjector(grad_dim=input_shape[-1],
                          proj_dim=proj_dim,
                          proj_type=proj_type,
                          seed=seed,
                          device='cuda:0',
                          dtype=dtype
                          )

    result = proj.project(g, model_id=0)
    result_again = proj.project(g, model_id=0)
    testing.assert_close(result, result_again, equal_nan=True)


@pytest.mark.parametrize("seed, proj_type, dtype, input_shape, proj_dim", PARAM)
@pytest.mark.cuda
def test_seed_consistency_2(seed,
                            proj_type,
                            dtype,
                            proj_dim,
                            input_shape,
                            ):
    """
    Check that re-initializing the class and re-running the same projection
    with the same seed leads to the same result.
    """

    g = testing.make_tensor(*input_shape, device='cuda:0', dtype=dtype)
    proj = BasicProjector(grad_dim=input_shape[-1],
                          proj_dim=proj_dim,
                          proj_type=proj_type,
                          seed=seed,
                          device='cuda:0',
                          dtype=dtype
                          )

    result = proj.project(g, model_id=0)

    proj_again = BasicProjector(grad_dim=input_shape[-1],
                                proj_dim=proj_dim,
                                proj_type=proj_type,
                                seed=seed,
                                device='cuda:0',
                                dtype=dtype)
    result_again = proj_again.project(g, model_id=0)
    testing.assert_close(result, result_again, equal_nan=True)


@pytest.mark.parametrize("seed, proj_type, dtype, input_shape, proj_dim", PARAM)
@pytest.mark.cuda
def test_norm_preservation(seed,
                           proj_type,
                           dtype,
                           proj_dim,
                           input_shape,
                           ):
    """
    Check that norms of differences are approximately preserved.
    """
    g = testing.make_tensor(*input_shape, device='cuda:0', dtype=dtype)
    proj = BasicProjector(grad_dim=input_shape[-1],
                          proj_dim=proj_dim,
                          proj_type=proj_type,
                          seed=seed,
                          device='cuda:0',
                          dtype=dtype
                          )

    p = proj.project(g, model_id=0)

    delta = 0.05
    eps = np.sqrt(np.log(1 / delta) / proj_dim)
    num_trials = 100
    num_successes = 0

    for _ in range(num_trials):
        i, j = np.random.choice(range(g.shape[0]), size=2)
        n = (g[i] - g[j]).norm()
        # assuming for the test that the norm of each column
        # of the projection matrix has norm sqrt(n)
        # (true for rademacher and approx true for gaussian)
        pn = (p[i] - p[j]).norm() / np.sqrt(input_shape[-1])
        res = (n - pn).cpu().abs().item()
        # 35 is an arbitrary constant
        # if NaN, just give up and count as success
        if math.isinf(res):
            print('aaaaaa')
        num_successes += int(res <= 35 * eps * n)
    assert num_successes >= num_trials * (1 - 3 * delta)  # leeway with 2 *


@pytest.mark.parametrize("seed, proj_type, dtype, input_shape, proj_dim", PARAM)
@pytest.mark.cuda
def test_prod_preservation(seed,
                           proj_type,
                           dtype,
                           proj_dim,
                           input_shape,
                           ):
    """
    Check that dot products are approximately preserved.
    """
    g = testing.make_tensor(*input_shape, device='cuda:0', dtype=dtype)
    proj = BasicProjector(grad_dim=input_shape[-1],
                          proj_dim=proj_dim,
                          proj_type=proj_type,
                          seed=seed,
                          device='cuda:0',
                          dtype=dtype
                          )

    # check that things break with a garbage matrix
    # (making sure the constant 15 is reasonable)
    # proj.proj_matrix = ch.empty_like(proj.proj_matrix)

    p = proj.project(g, model_id=0)

    delta = 0.2
    eps = np.sqrt(np.log(1 / delta) / proj_dim)
    num_trials = 100
    num_successes = 0

    for _ in range(num_trials):
        i, j = np.random.choice(range(g.shape[0]), size=2)
        n = (g[i] @ g[j])
        pn = ((p[i] / np.sqrt(input_shape[-1])) @ (p[j] / input_shape[-1]))
        res = (n.abs() - pn.abs()).cpu().abs().item()
        t = (50 * np.sqrt(proj_dim) * eps * n).abs().item()
        # if NaN, just give up and count as success
        num_successes += max(int(res <= t), math.isinf(res), math.isinf(t))

    assert num_successes >= num_trials * (1 - 2 * delta)


@pytest.mark.parametrize("seed, proj_type, dtype, input_shape, proj_dim", PARAM)
@pytest.mark.cuda
def test_single_nonzero_feature(seed,
                                proj_type,
                                dtype,
                                proj_dim,
                                input_shape,
                                ):
    """
    Check that output takes into account every feature.
    """
    g = ch.zeros(*input_shape, device='cuda:0', dtype=dtype)
    for ind in range(input_shape[0]):
        coord = np.random.choice(range(input_shape[1]))
        val = ch.randn(1)
        g[ind, coord] = val.item()

    proj = BasicProjector(grad_dim=input_shape[-1],
                          proj_dim=proj_dim,
                          proj_type=proj_type,
                          seed=seed,
                          device='cuda:0',
                          dtype=dtype
                          )
    p = proj.project(g, model_id=0)
    assert (~ch.isclose(p, ch.zeros_like(p))).all().item()


@pytest.mark.parametrize("seed, proj_type, dtype, input_shape, proj_dim", PARAM)
@pytest.mark.cuda
def test_first_nonzero_feature(seed,
                               proj_type,
                               dtype,
                               proj_dim,
                               input_shape,
                               ):
    """
    Check that output takes into account first features.
    """
    g = ch.zeros(*input_shape, device='cuda:0', dtype=dtype)
    g[:, 0] = 1.

    proj = BasicProjector(grad_dim=input_shape[-1],
                          proj_dim=proj_dim,
                          proj_type=proj_type,
                          seed=seed,
                          device='cuda:0',
                          dtype=dtype
                          )
    p = proj.project(g, model_id=0)
    assert (~ch.isclose(p, ch.zeros_like(p))).all().item()


@pytest.mark.parametrize("seed, proj_type, dtype, input_shape, proj_dim", PARAM)
@pytest.mark.cuda
def test_last_nonzero_feature(seed,
                              proj_type,
                              dtype,
                              proj_dim,
                              input_shape,
                              ):
    """
    Check that output takes into account last features.
    """
    g = ch.zeros(*input_shape, device='cuda:0', dtype=dtype)
    g[:, -1] = 1.

    proj = BasicProjector(grad_dim=input_shape[-1],
                          proj_dim=proj_dim,
                          proj_type=proj_type,
                          seed=seed,
                          device='cuda:0',
                          dtype=dtype
                          )
    p = proj.project(g, model_id=0)
    assert (~ch.isclose(p, ch.zeros_like(p))).all().item()


@pytest.mark.parametrize("seed, proj_type, dtype, input_shape, proj_dim", PARAM)
@pytest.mark.cuda
def test_same_features(seed,
                       proj_type,
                       dtype,
                       proj_dim,
                       input_shape,
                       ):
    """
    Check that output is the same for the same features
    """
    g = testing.make_tensor(*input_shape, device='cuda:0', dtype=dtype)
    g[-1] = g[0]

    proj = BasicProjector(grad_dim=input_shape[-1],
                          proj_dim=proj_dim,
                          proj_type=proj_type,
                          seed=seed,
                          device='cuda:0',
                          dtype=dtype
                          )
    p = proj.project(g, model_id=0)

    assert ch.allclose(p[0], p[-1])


@pytest.mark.parametrize("seed, proj_type, dtype, input_shape, proj_dim", PARAM)
@pytest.mark.cuda
def test_orthogonality(seed,
                       proj_type,
                       dtype,
                       proj_dim,
                       input_shape,
                       ):
    """
    Check that orthgonality of inputs is approximately presereved whp
    """
    if input_shape[0] == 1:
        pass
    else:
        proj = BasicProjector(grad_dim=input_shape[-1],
                              proj_dim=proj_dim,
                              proj_type=proj_type,
                              seed=seed,
                              device='cuda:0',
                              dtype=dtype)

        num_successes = 0
        num_trials = 100
        for _ in range(num_trials):
            g = testing.make_tensor(*input_shape, device='cuda:0', dtype=dtype)
            g[-1] -= g[0] @ g[-1] / (g[0].norm() ** 2) * g[0]
            p = proj.project(g, model_id=0)
            if p[0] @ p[-1] < 1e-3:
                num_successes += 1
        assert num_successes > 0.33 * num_trials



from tqdm import tqdm
from pathlib import Path
from itertools import product
import pytest
import torch as ch

from trak import TRAKer
from trak.projectors import BasicProjector

from .utils import construct_rn9, get_dataloader, eval_correlations
from .utils import download_cifar_checkpoints, download_cifar_betons


def get_projector(use_cuda_projector, dtype):
    if use_cuda_projector:
        return None
    return BasicProjector(grad_dim=2273856, proj_dim=1024,
                          seed=0, proj_type='rademacher',
                          dtype=dtype, device='cuda:0')


PARAM = list(product([False, True],  # serialize
                     [False, True],  # basic / cuda projector
                     [ch.float16, ch.float32],  # projection dtype
                     [100, 32],  # batch size
                     ))


@pytest.mark.parametrize("serialize, use_cuda_projector, dtype, batch_size", PARAM)
@pytest.mark.cuda
def test_cifar_acc(serialize, use_cuda_projector, dtype, batch_size, tmp_path):
    device = 'cuda:0'
    projector = get_projector(use_cuda_projector, dtype)
    model = construct_rn9().to(memory_format=ch.channels_last).to(device)
    model = model.eval()

    BETONS_PATH = Path(tmp_path).joinpath('cifar_betons')
    BETONS = download_cifar_betons(BETONS_PATH)

    loader_train = get_dataloader(BETONS, batch_size=batch_size, split='train')
    loader_val = get_dataloader(BETONS, batch_size=batch_size, split='val')

    CKPT_PATH = Path(tmp_path).joinpath('cifar_ckpts')
    ckpt_files = download_cifar_checkpoints(CKPT_PATH)
    ckpts = [ch.load(ckpt, map_location='cpu') for ckpt in ckpt_files]

    traker = TRAKer(model=model,
                    task='image_classification',
                    projector=projector,
                    proj_dim=1024,
                    train_set_size=10_000,
                    save_dir=tmp_path,
                    device=device)

    for model_id, ckpt in enumerate(ckpts):
        traker.load_checkpoint(checkpoint=ckpt, model_id=model_id)
        for batch in tqdm(loader_train, desc='Computing TRAK embeddings...'):
            traker.featurize(batch=batch, num_samples=len(batch[0]))

    traker.finalize_features()

    if serialize:
        del traker
        traker = TRAKer(model=model,
                        task='image_classification',
                        projector=projector,
                        proj_dim=1024,
                        train_set_size=10_000,
                        save_dir=tmp_path,
                        device=device)

    for model_id, ckpt in enumerate(ckpts):
        traker.start_scoring_checkpoint(ckpt, model_id, num_targets=2_000)
        for batch in tqdm(loader_val, desc='Scoring...'):
            traker.score(batch=batch, num_samples=len(batch[0]))

    scores = traker.finalize_scores().cpu()

    avg_corr = eval_correlations(infls=scores, tmp_path=tmp_path)
    assert avg_corr > 0.062, 'correlation with 3 CIFAR-2 models should be >= 0.062'



from torch import Tensor
import torch
ch = torch


def parameters_to_vector(parameters) -> Tensor:
    """
    Same as https://pytorch.org/docs/stable/generated/torch.nn.utils.parameters_to_vector.html
    but with :code:`reshape` instead of :code:`view` to avoid a pesky error.
    """
    vec = []
    for param in parameters:
        vec.append(param.reshape(-1))
    return ch.cat(vec)


def get_num_params(model: torch.nn.Module) -> int:
    return parameters_to_vector(model.parameters()).numel()


def get_params_dict(model: torch.nn.Module) -> list:
    return [x[0] for x in list(model.named_parameters())]


def is_not_buffer(ind, params_dict) -> bool:
    name = params_dict[ind]
    if ('running_mean' in name) or ('running_var' in name) or ('num_batches_tracked' in name):
        return False
    return True


def vectorize_and_ignore_buffers(g, params_dict=None) -> Tensor:
    """
    gradients are given as a tuple :code:`(grad_w0, grad_w1, ... grad_wp)` where
    :code:`p` is the number of weight matrices. each :code:`grad_wi` has shape
    :code:`[batch_size, ...]` this function flattens :code:`g` to have shape
    :code:`[batch_size, num_params]`.
    """
    batch_size = len(g[0])
    out = []
    if params_dict is not None:
        for b in range(batch_size):
            out.append(ch.cat([x[b].flatten() for i, x in enumerate(g) if is_not_buffer(i, params_dict)]))
    else:
        for b in range(batch_size):
            out.append(ch.cat([x[b].flatten() for x in g]))
    return ch.stack(out)



from abc import ABC, abstractmethod
from typing import Iterable, Optional
from torch import Tensor
from .utils import parameters_to_vector, vectorize_and_ignore_buffers, get_params_dict
from .modelout_functions import AbstractModelOutput
import torch
ch = torch
try:
    from functorch import grad, vmap, make_functional_with_buffers
except ImportError:
    print('Cannot import `functorch`. Functional mode cannot be used.')


class AbstractGradientComputer(ABC):
    """ Implementations of the GradientComputer class should allow for
    per-sample gradients.  This is behavior is enabled with three methods:

    - the :meth:`.load_model_params` method, well, loads model parameters. It can
      be as simple as a :code:`self.model.load_state_dict(..)`

    - the :meth:`.compute_per_sample_grad` method computes per-sample gradients
      of the chosen model output function with respect to the model's parameters.

    - the :meth:`.compute_loss_grad` method computes the gradients of the loss
      function with respect to the model output (which should be a scalar) for
      every sample.

    The class attribute :code:`is_functional` is used to determine what
    implementation of ModelOutput to use, i.e. whether it should use
    :code:`functorch`'s functional models.
    """
    is_functional = True

    @abstractmethod
    def __init__(self,
                 model: torch.nn.Module,
                 modelout_fn: AbstractModelOutput,
                 grad_dim: Optional[int] = None,
                 ) -> None:
        """ Initializes attributes, nothing too interesting happening.

        Args:
            model (torch.nn.Module):
                model
            modelout_fn (AbstractModelOutput):
                model output function
            grad_dim (int, optional):
                Size of the gradients (number of model parameters). Defaults to
                None.

        """
        self.model = model
        self.modelout_fn = modelout_fn()
        self.grad_dim = grad_dim

    @abstractmethod
    def load_model_params(self, model) -> None:
        ...

    @abstractmethod
    def compute_per_sample_grad(self, batch: Iterable[Tensor], batch_size: int) -> Tensor:
        ...

    @abstractmethod
    def compute_loss_grad(self, batch: Iterable[Tensor], batch_size: int) -> Tensor:
        ...


class FunctionalGradientComputer(AbstractGradientComputer):
    def __init__(self,
                 model: torch.nn.Module,
                 modelout_fn: AbstractModelOutput,
                 grad_dim: int) -> None:
        super().__init__(model, modelout_fn, grad_dim)
        self.load_model_params(model)

    def load_model_params(self, model) -> None:
        """ Given a a torch.nn.Module model, inits/updates the func_model, along
        with its weights and buffers. See
        https://pytorch.org/functorch/stable/generated/functorch.make_functional_with_buffers.html#functorch-make-functional-with-buffers
        for more details on `functorch`'s functional models.

        Args:
            model (torch.nn.Module):
                model to load

        """
        self.func_model, self.func_weights, self.func_buffers = make_functional_with_buffers(model)
        self.params_dict = get_params_dict(model)

    def compute_per_sample_grad(self,
                                batch: Iterable[Tensor],
                                batch_size: Optional[int] = None,
                                ) -> Tensor:
        """ Uses functorch's :code:`vmap` (see
        https://pytorch.org/functorch/stable/generated/functorch.vmap.html#functorch.vmap
        for more details) to vectorize the computations of per-sample gradients.

        Doesn't use :code:`batch_size`; only added to follow the abstract method
        signature.

        Args:
            batch (Iterable[Tensor]):
                batch of data
            batch_size (int, optional):
                Defaults to None.

        Returns:
            Tensor:
                gradients of the model output function of each sample in the
                batch with respect to the model's parameters.

        """
        # taking the gradient wrt weights (second argument of get_output, hence argnums=1)
        grads_loss = grad(self.modelout_fn.get_output, has_aux=False, argnums=1)
        # map over batch dimensions (hence 0 for each batch dimension, and None for model params)
        grads = vmap(grads_loss,
                     in_dims=(None, None, None, *([0] * len(batch))),
                     randomness='different')(self.func_model, self.func_weights,
                                             self.func_buffers, *batch)
        return vectorize_and_ignore_buffers(grads, self.params_dict)

    def compute_loss_grad(self, batch: Iterable[Tensor]) -> Tensor:
        """Computes the gradient of the loss with respect to the model output

        .. math::

            \\partial \\ell / \\partial \\text{(model output)}

        Note: For all applications we considered, we analytically derived the
        out-to-loss gradient, thus avoiding the need to do any backward passes
        (let alone per-sample grads). If for your application this is not feasible,
        you'll need to subclass this and modify this method to have a structure
        similar to the one of :meth:`FunctionalGradientComputer:.get_output`,
        i.e. something like:

        .. code-block:: python

            grad_out_to_loss = grad(self.model_out_to_loss_grad, ...)
            grads = vmap(grad_out_to_loss, ...)
            ...

        Args:
            batch (Iterable[Tensor]):
                batch of data

        """
        return self.modelout_fn.get_out_to_loss_grad(self.func_model,
                                                     self.func_weights,
                                                     self.func_buffers,
                                                     batch)


class IterativeGradientComputer(AbstractGradientComputer):
    is_functional = False

    def __init__(self,
                 model,
                 modelout_fn: AbstractModelOutput,
                 grad_dim: int) -> None:
        super().__init__(model, modelout_fn, grad_dim)
        self.load_model_params(model)

    def load_model_params(self, model) -> Tensor:
        self.model = model
        self.model_params = list(self.model.parameters())

    def compute_per_sample_grad(self,
                                batch: Iterable[Tensor],
                                batch_size: int,
                                ) -> Tensor:
        """ Computes per-sample gradients of the model output function This
        method does not leverage vectorization (and is hence much slower than
        its equivalent in :class:`.FunctionalGradientComputer`). We recommend
        that you use this only if :code:`functorch` is not available to you,
        e.g. you have a (very) old version of pytorch.

        Args:
            batch (Iterable[Tensor]):
                batch of data
            batch_size (int, optional):
                Defaults to None.

        Returns:
            Tensor:
                gradients of the model output function of each sample in the
                batch with respect to the model's parameters.

        """
        grads = ch.zeros(batch_size, self.grad_dim).to(batch[0].device)

        margin = self.modelout_fn.get_output(self.model, *batch)
        for ind in range(batch_size):
            grads[ind] = parameters_to_vector(ch.autograd.grad(margin[ind],
                                                               self.model_params,
                                                               retain_graph=True))
        return grads

    def compute_loss_grad(self, batch: Iterable[Tensor]) -> Tensor:
        """Computes the gradient of the loss with respect to the model output

        .. math::

            \\partial \\ell / \\partial \\text{(model output)}

        Note: For all applications we considered, we analytically derived the
        out-to-loss gradient, thus avoiding the need to do any backward passes
        (let alone per-sample grads). If for your application this is not feasible,
        you'll need to subclass this and modify this method to have a structure
        similar to the one of :meth:`.IterativeGradientComputer.get_output`,
        i.e. something like:

        .. code-block:: python

            out_to_loss = self.model_out_to_loss(...)
            for ind in range(batch_size):
                grads[ind] = torch.autograd.grad(out_to_loss[ind], ...)
            ...

        Args:
            batch (Iterable[Tensor]):
                batch of data

        """
        return self.modelout_fn.get_out_to_loss_grad(self.model, batch)



import pytest
from tqdm import tqdm
from torch.utils.data import DataLoader
from torchvision import datasets, models, transforms

from trak import TRAKer
from trak.projectors import BasicProjector
from trak.gradient_computers import IterativeGradientComputer


def test_cifar10(tmp_path, device='cpu'):
    model = models.resnet18(weights='DEFAULT')
    model.to(device)
    model.eval()

    transform = transforms.Compose([transforms.ToTensor(),
                                    transforms.Normalize((0.5, 0.5, 0.5), (0.5, 0.5, 0.5))])

    ds_train = datasets.CIFAR10(root='/tmp', download=True, train=True, transform=transform)
    loader_train = DataLoader(ds_train, batch_size=10, shuffle=False)
    if device == 'cpu':
        # the default CudaProjector does not work on cpu
        projector = BasicProjector(grad_dim=sum(x.numel() for x in model.parameters()),
                                   proj_dim=20,
                                   seed=0,
                                   proj_type='rademacher',
                                   device=device)
    else:
        projector = None
    traker = TRAKer(model=model,
                    task='image_classification',
                    train_set_size=len(ds_train),
                    projector=projector,
                    save_dir=tmp_path,
                    device=device)

    ckpts = [model.state_dict(), model.state_dict()]
    for model_id, ckpt in enumerate(ckpts):
        traker.load_checkpoint(ckpt, model_id=model_id)

        for batch in tqdm(loader_train, desc='Computing TRAK embeddings...'):
            batch = [x.to(device) for x in batch]
            traker.featurize(batch=batch,
                             num_samples=loader_train.batch_size)
            break  # a CPU pass takes too long lol

    traker.finalize_features()

    ds_val = datasets.CIFAR10(root='/tmp', download=True, train=False, transform=transform)
    loader_val = DataLoader(ds_val, batch_size=10, shuffle=False)
    for model_id, ckpt in enumerate(ckpts):
        traker.start_scoring_checkpoint(ckpt, model_id, num_targets=100)
        for batch in tqdm(loader_val, desc='Scoring...'):
            batch = [x.to(device) for x in batch]
            traker.score(batch=batch,
                         num_samples=loader_val.batch_size)
            break  # a CPU pass takes too long lol

    traker.finalize_scores()


@pytest.mark.cuda
def test_cifar10_cuda(tmp_path):
    test_cifar10(tmp_path, device='cuda:0')


def test_cifar10_iter(tmp_path, device='cpu'):
    model = models.resnet18(weights='DEFAULT')
    model.to(device)
    model.eval()

    transform = transforms.Compose([transforms.ToTensor(),
                                    transforms.Normalize((0.5, 0.5, 0.5), (0.5, 0.5, 0.5))])

    ds_train = datasets.CIFAR10(root='/tmp', download=True, train=True, transform=transform)
    loader_train = DataLoader(ds_train, batch_size=10, shuffle=False)

    if device == 'cpu':
        # the default CudaProjector does not work on cpu
        projector = BasicProjector(grad_dim=sum(x.numel() for x in model.parameters()),
                                   proj_dim=20,
                                   seed=0,
                                   proj_type='rademacher',
                                   device=device)
    else:
        projector = None

    traker = TRAKer(model=model,
                    task='image_classification',
                    train_set_size=len(ds_train),
                    save_dir=tmp_path,
                    projector=projector,
                    gradient_computer=IterativeGradientComputer,
                    device=device)

    traker.load_checkpoint(model.state_dict(), model_id=0)
    for batch in tqdm(loader_train, desc='Computing TRAK embeddings...'):
        batch = [x.to(device) for x in batch]
        traker.featurize(batch=batch,
                         num_samples=loader_train.batch_size)
        break  # a CPU pass takes too long lol

    traker.finalize_features()

    ds_val = datasets.CIFAR10(root='/tmp', download=True, train=False, transform=transform)
    loader_val = DataLoader(ds_val, batch_size=10, shuffle=False)
    # load margins
    traker.start_scoring_checkpoint(model.state_dict(), model_id=0, num_targets=100)
    for batch in tqdm(loader_val, desc='Scoring...'):
        batch = [x.to(device) for x in batch]
        traker.score(batch=batch, num_samples=loader_val.batch_size)
        break


@pytest.mark.cuda
def test_cifar10_iter_cuda(tmp_path):
    test_cifar10_iter(tmp_path, device='cuda:0')



from abc import ABC, abstractmethod
from typing import Optional, Iterable, Union
from pathlib import Path
import os
import uuid
import json
import numpy as np
from numpy.lib.format import open_memmap
import torch
ch = torch


class AbstractSaver(ABC):
    """
    Implementations of Saver class must implement getters and setters for TRAK
    features and scores, as well as intermediate values like gradients and
    "out-to-loss-gradient".
    
    The Saver class also handles the recording of metadata associated with each
    TRAK run. For example, hyperparameters like "JL dimension" -- the dimension
    used for the dimensionality reduction step of TRAK (Johnson-Lindenstrauss
    projection).
    """
    @abstractmethod
    def __init__(self,
                 save_dir: Union[Path, str],
                 metadata: Iterable,
                 load_from_save_dir: bool) -> None:
        """ Creates the save directory if it doesn't already exist.
        If the save directory already exists, it validates that the current
        TRAKer class has the same hyperparameters (metadata) as the one
        specified in the save directory. Next, this method loads any existing
        computed results / intermediate values in the save directory.
        Last, it initalizes the self.current_* attributes which will be later
        populated with data for the "current" model ID of the TRAKer instance.

        Args:
            save_dir (Union[Path, str]): directory to save TRAK results,
                intermediate values, and metadata
            metadata (Iterable): a dictionary containing metadata related to the
                TRAKer class
            load_from_save_dir (bool): If True, the Saver instance will attempt
                to load existing metadata from save_dir. May lead to I/O issues
                if multiple Saver instances ran in parallel have this flag set
                to True. See the SLURM tutorial for more details.
        """
        self.metadata = metadata
        self.save_dir = Path(save_dir).resolve()
        self.load_from_save_dir = load_from_save_dir
        os.makedirs(self.save_dir, exist_ok=True)

        # init TRAKer metadata
        self.metadata_file = self.save_dir.joinpath('metadata.json')
        if os.path.exists(self.metadata_file) and self.load_from_save_dir:
            with open(self.metadata_file, 'r') as f:
                existsing_metadata = json.load(f)
            existing_jl_dim = int(existsing_metadata['JL dimension'])
            assert self.metadata['JL dimension'] == existing_jl_dim,\
                   f"In {self.save_dir} there are models using JL dimension {existing_jl_dim}\
                   , and this TRAKer instance uses JL dimension {self.metadata['JL dimension']}."

            existing_matrix_type = existsing_metadata['JL matrix type']
            assert self.metadata['JL matrix type'] == existing_matrix_type,\
                   f"In {self.save_dir} there are models using a {existing_matrix_type} JL matrix\
                   , and this TRAKer instance uses a {self.metadata['JL matrix type']} JL matrix."

        elif self.load_from_save_dir:
            with open(self.metadata_file, 'w') as f:
                json.dump(self.metadata, f)

        self.model_ids = {}
        if self.load_from_save_dir:
            # check if there are existing model ids in the save_dir
            self.model_ids_files = self.save_dir.rglob('id_*.json')

            for existing_model_id_file in self.model_ids_files:
                with open(existing_model_id_file, 'r') as f:
                    existing_id = json.load(f)
                    existing_id = {int(model_id): metadata
                                   for model_id, metadata in existing_id.items()}
                self.model_ids.update(existing_id)

        # wlog set num_targets to those of a random model_id we could raise an
        # error here if different model_ids have different num_targets but this
        # could be a bit too stringent in some cases
        if len(self.model_ids.keys()) > 0:
            self.num_targets = list(self.model_ids.values())[0]['targets_scored']

        print(f'Existing IDs in {self.save_dir}: {list(self.model_ids.keys())}')

        self.current_model_id = None
        self.current_grads = None
        self.current_out_to_loss = None
        self.current_features = None
        self.current_target_grads = None

    @abstractmethod
    def register_model_id(self, model_id: int) -> None:
        """ Create metadata for a new model ID (checkpoint).

        Args:
            model_id (int):
                a unique ID for a checkpoint

        """
        ...

    @abstractmethod
    def serialize_model_id_metadata(self, model_id: int) -> None:
        """ Write to disk / commit any updates to the metadata associated
        to a given model ID

        Args:
            model_id (int):
                a unique ID for a checkpoint

        """
        ...

    @abstractmethod
    def load_store(self, model_id: int) -> None:
        """ Populates the self.current_* attributes with data for the
        given model ID (checkpoint).

        Args:
            model_id (int):
                a unique ID for a checkpoint

        """
        ...

    @abstractmethod
    def load_target_store(self,
                          model_id: int,
                          num_targets: int,
                          mode: Optional[str]) -> None:
        """ Loads/initializes metadata and store arrays for target results and
        features (gradients)

        Args:
            model_id (int):
                a unique ID for a checkpoint
            num_targets (int):
                number of target samples that will be scored

        """
        ...

    @abstractmethod
    def del_grads(self, model_id: int, target: bool) -> None:
        """ Delete the intermediate values (gradients) for a given model id

        Args:
            model_id (int):
                a unique ID for a checkpoint
            target (bool):
                if True, delete the gradients of the target samples, otherwise
                delete the train set gradients.

        """
        ...

    @abstractmethod
    def clear_target_grad_count(self, model_id: int) -> None:
        """ Reset the count for how many targets we are scoring

        Args:
            model_id (int):
                a unique ID for a checkpoint

        """
        ...


class ModelIDException(Exception):
    """ A minimal custom exception for errors related to model IDs """
    pass


class MmapSaver(AbstractSaver):
    """ A saver that uses memory-mapped numpy arrays. This makes small reads and
    writes (e.g.) during featurizing feasible without loading the entire file
    into memory.

    """
    def __init__(self, save_dir, metadata, train_set_size, proj_dim, load_from_save_dir) -> None:
        super().__init__(save_dir=save_dir,
                         metadata=metadata,
                         load_from_save_dir=load_from_save_dir)
        self.train_set_size = train_set_size
        self.proj_dim = proj_dim

    def register_model_id(self,
                          model_id: int,
                          _allow_featurizing_already_registered: bool) -> None:
        """ This method
        1) checks if the model ID already exists in the save dir
        2) if yes, it raises an error since model IDs must be unique
        3) if not, it creates a metadata file for it and initalizes store mmaps

        Args:
            model_id (int):
                a unique ID for a checkpoint

        Raises:
            ModelIDException:
                raised if the model ID to be registered already exists

        """
        self.current_model_id = model_id

        if self.current_model_id in self.model_ids.keys() and (not _allow_featurizing_already_registered):
            err_msg = f'model id {self.current_model_id} is already registered. Check {self.save_dir}'
            raise ModelIDException(err_msg)
        self.model_ids[self.current_model_id] = {'featurized': 0,
                                                 'finalized': 0,
                                                 'targets_scored': 0}

        self.init_store(self.current_model_id)
        self.serialize_model_id_metadata(self.current_model_id)

    def serialize_model_id_metadata(self, model_id) -> None:
        with open(self.save_dir.joinpath(f'id_{model_id}.json'), 'w+') as f:
            content = {self.current_model_id: self.model_ids[self.current_model_id]}
            json.dump(content, f)

    def init_store(self, model_id) -> None:
        prefix = self.save_dir.joinpath(str(model_id))
        try:
            os.makedirs(prefix)
        except FileExistsError:
            raise ModelIDException(f'model id folder {prefix} already exists')

        self.load_store(model_id, mode='w+')

    def load_store(self,
                   model_id: int,
                   mode: Optional[str] = 'r+') -> None:
        """ This method uses numpy memmaps for serializing the TRAK results and
        intermediate values.

        Args:
            model_id (int):
                a unique ID for a checkpoint
            mode (str, optional):
                Defaults to 'r+'.

        """
        self.current_model_id = model_id
        prefix = self.save_dir.joinpath(str(model_id))

        self.current_grads_path = prefix.joinpath('grads.mmap')
        self.current_grads = open_memmap(filename=self.current_grads_path,
                                         mode=mode,
                                         shape=(self.train_set_size, self.proj_dim),
                                         dtype=np.float32)

        self.current_out_to_loss_path = prefix.joinpath('out_to_loss.mmap')
        self.current_out_to_loss = open_memmap(filename=self.current_out_to_loss_path,
                                               mode=mode,
                                               shape=(self.train_set_size, 1),
                                               dtype=np.float32)

        self.current_features_path = prefix.joinpath('features.mmap')
        self.current_features = open_memmap(filename=self.current_features_path,
                                            mode=mode,
                                            shape=(self.train_set_size, self.proj_dim),
                                            dtype=np.float32)

    def load_target_store(self,
                          model_id: int,
                          num_targets: int,
                          mode: Optional[str] = 'r+',
                          ) -> None:
        """ This method uses numpy memmaps for serializing the TRAK intermediate
        values for targets.

        Args:
            model_id (int):
                a unique ID for a checkpoint
            num_targets (int):
                number of target samples that will be scored

        """
        self.num_targets = num_targets
        self.current_model_id = model_id
        self.model_ids[model_id]['targets_scored'] = self.num_targets
        self.serialize_model_id_metadata(model_id)

        prefix = self.save_dir.joinpath(str(model_id))

        self.current_target_grads_path = prefix.joinpath('grads_target.mmap')
        self.current_target_grads = open_memmap(filename=self.current_target_grads_path,
                                                mode=mode,
                                                shape=(num_targets, self.proj_dim),
                                                dtype=np.float32)

    def save_scores(self, scores, exp_name):
        prefix = self.save_dir.joinpath('scores')
        if exp_name is None:
            exp_name = str(uuid.uuid4())  # generate random unique ID
            print(f'Saving scores in {prefix}/scores_{exp_name}.npy')
        filename = 'scores_' + exp_name + '.npy'
        if not os.path.isdir(prefix):
            os.makedirs(prefix)
        np.save(prefix.joinpath(filename), scores)

    def del_grads(self, model_id, target=False):
        if target:
            grads_file = self.save_dir.joinpath(str(model_id)).joinpath('grads_target.mmap')
            self.clear_target_grad_count(model_id)
        else:
            grads_file = self.save_dir.joinpath(str(model_id)).joinpath('grads.mmap')

        # delete grads memmap
        grads_file.unlink()

    def clear_target_grad_count(self, model_id):
        self.model_ids[model_id]['targets_scored'] = 0
        self.serialize_model_id_metadata(model_id)

        if model_id == self.current_model_id:
            self.current_num_target_grads = 0



"""
Here we provide an abstract "model output" class AbstractModelOutput, together with a number
of subclasses for particular applications (vision, language, etc):

- :class:`.ImageClassificationModelOutput`
- :class:`.IterImageClassificationModelOutput`
- :class:`.CLIPModelOutput`
- :class:`.TextClassificationModelOutput`

These classes implement methods that transform input batches to the desired model output
(e.g. logits, loss, etc).
See Sections 2 & 3 of `our paper <https://arxiv.org/abs/2303.14186>`_ for more
details on what model output functions are in the context of TRAK and how to use
& design them.

See, e.g. `this tutorial <https://trak.readthedocs.io/en/latest/clip.html>`_ for an
example on how to subclass :code:`AbstractModelOutput` for a task of your
choice.
"""
from abc import ABC, abstractmethod
from typing import Iterable
from torch import Tensor
from torch.nn import Module
import torch as ch


class AbstractModelOutput(ABC):
    """ See, e.g. `this tutorial <https://trak.readthedocs.io/en/latest/clip.html>`_
    for an example on how to subclass :code:`AbstractModelOutput` for a task of
    your choice.

    Subclasses must implement:

    - a :code:`get_output` method that takes in a batch of inputs and model
      weights to produce outputs that TRAK will be trained to predict. In the
      notation of the paper, :code:`get_output` should return :math:`f(z,\\theta)`

    - a :code:`get_out_to_loss_grad` method that takes in a batch of inputs and
      model weights to produce the gradient of the function that transforms the
      model outputs above into the loss with respect to the batch. In the
      notation of the paper, :code:`get_out_to_loss_grad` returns (entries along
      the diagonal of) :math:`Q`.

    """
    @abstractmethod
    def __init__(self) -> None:
        pass

    @abstractmethod
    def get_output(self,
                   model,
                   batch: Iterable[Tensor]) -> Tensor:
        """ See Sections 2 & 3 of `our paper
        <https://arxiv.org/abs/2303.14186>`_ for more details on what model
        output functions are in the context of TRAK and how to use & design
        them.

        Args:
            model (torch.nn.Module):
                model
            batch (Iterable[Tensor]):
                input batch

        Returns:
            Tensor:
                model output function
        """
        ...

    @abstractmethod
    def get_out_to_loss_grad(self,
                             model,
                             batch: Iterable[Tensor]) -> Tensor:
        """ See Sections 2 & 3 of `our paper
        <https://arxiv.org/abs/2303.14186>`_ for more details on what the
        out-to-loss functions (in the notation of the paper, :math:`Q`) are in
        the context of TRAK and how to use & design them.

        Args:
            model (torch.nn.Module): model
            batch (Iterable[Tensor]): input batch

        Returns:
            Tensor: gradient of the out-to-loss function
        """
        ...


class ImageClassificationModelOutput(AbstractModelOutput):
    """ Margin for (multiclass) image classification. See Section 3.3 of `our
    paper <https://arxiv.org/abs/2303.14186>`_ for more details.
    """

    def __init__(self, temperature: float = 1.) -> None:
        """
        Args:
            temperature (float, optional): Temperature to use inside the
            softmax for the out-to-loss function. Defaults to 1.
        """
        super().__init__()
        self.softmax = ch.nn.Softmax(-1)
        self.loss_temperature = temperature

    @staticmethod
    def get_output(func_model,
                   weights: Iterable[Tensor],
                   buffers: Iterable[Tensor],
                   image: Tensor,
                   label: Tensor) -> Tensor:
        """ For a given input :math:`z=(x, y)` and model parameters :math:`\\theta`,
        let :math:`p(z, \\theta)` be the softmax probability of the correct class.
        This method implements the model output function

        .. math::

            \\log(\\frac{p(z, \\theta)}{1 - p(z, \\theta)}).

        It uses functorch's functional models to make the per-sample gradient computations faster.
        For more details on what func_model, weights & buffers are, and how to use them, please
        refer to https://pytorch.org/functorch/stable/ and
        https://pytorch.org/functorch/stable/notebooks/per_sample_grads.html.

        Note: this method slightly breaks abstraction since it has a different
        signature from the abstract :code:`get_output`. It's only used
        internally by :func:`GradientComputer` so it shouldn't be a big problem.
        If you're reading this and feel strongly about it, feel free to make a
        PR with a fix.

        Args:
            func_model (func):
                functorch functional model
            weights (Iterable[Tensor]):
                functorch model weights
            buffers (Iterable[Tensor]):
                functorch model buffers
            image (Tensor):
                input image, should not have batch dimension
            label (Tensor):
                input label, should not have batch dimension

        Returns:
            Tensor:
                model output for the given image-label pair :math:`z` and
                weights & buffers :math:`\\theta`.
        """
        logits = func_model(weights, buffers, image.unsqueeze(0))
        bindex = ch.arange(logits.shape[0]).to(logits.device, non_blocking=False)
        logits_correct = logits[bindex, label.unsqueeze(0)]

        cloned_logits = logits.clone()
        # a hacky way to remove the logits of the correct labels from the sum
        # in logsumexp by setting to -ch.inf
        cloned_logits[bindex, label.unsqueeze(0)] = ch.tensor(-ch.inf).to(logits.device)

        margins = logits_correct - cloned_logits.logsumexp(dim=-1)
        return margins.sum()

    def forward(self, model: Module, batch: Iterable[Tensor]) -> Tensor:
        """ Utility method to make forward passes compatible across different models,
        e.g. image classification models take in only the images in the batch,
        but CLIP takes in both the images and captions --- using this method,
        the TRAKer class does not need to be modified when we have a new model
        that uses different parts of the input batch.

        Args:
            model (Module):
                model
            batch (Iterable[Tensor]):
                input batch

        Returns:
            Tensor:
                model output (not to be confused with the model output function)
        """
        images, _ = batch
        return model(images)

    def get_out_to_loss_grad(self, func_model, weights, buffers, batch: Iterable[Tensor]) -> Tensor:
        """ Computes the (reweighting term Q in the paper)

        Args:
            func_model (func):
                functorch functional model
            weights (Iterable[Tensor]):
                functorch model weights
            buffers (Iterable[Tensor]):
                functorch model buffers
            batch (Iterable[Tensor]):
                input batch

        Returns:
            Tensor:
                out-to-loss (reweighting term) for the input batch
        """
        images, labels = batch
        logits = func_model(weights, buffers, images)
        # here we are directly implementing the gradient instead of relying on autodiff to do
        # that for us
        ps = self.softmax(logits / self.loss_temperature)[ch.arange(logits.size(0)), labels]
        return (1 - ps).clone().detach().unsqueeze(-1)


class IterImageClassificationModelOutput(AbstractModelOutput):
    """
    Margin for (multiclass) image classification. See Section 3.3 of `our paper
    <https://arxiv.org/abs/2303.14186>`_ for more details.
    """

    def __init__(self, temperature=1.) -> None:
        """
        Args:
            temperature (float, optional):
                Temperature to use inside the softmax for the out-to-loss
                function. Defaults to 1.
        """
        super().__init__()
        self.softmax = ch.nn.Softmax(-1)
        self.loss_temperature = temperature

    def get_output(self,
                   model: Module,
                   images: Tensor,
                   labels: Tensor) -> Tensor:
        """ For a given input :math:`z=(x, y)` and model parameters :math:`\\theta`,
        let :math:`p(z, \\theta)` be the softmax probability of the correct class.
        This method implements the model output function

        .. math::

            \\log(\\frac{p(z, \\theta)}{1 - p(z, \\theta)}).

        Args:
            model (torch.nn.Module):
                model
            images (Tensor):
                input images
            labels (Tensor):
                input labels

        Returns:
            Tensor:
                model output for the given image-label pair :math:`z` and model
                parameters :math:`\\theta`.
        """
        logits = model(images)
        bindex = ch.arange(logits.shape[0]).to(logits.device, non_blocking=False)
        logits_correct = logits[bindex, labels]

        cloned_logits = logits.clone()
        # a hacky way to remove the logits of the correct labels from the sum
        # in logsumexp by setting to -ch.inf
        cloned_logits[bindex, labels] = ch.tensor(-ch.inf).to(logits.device)

        margins = logits_correct - cloned_logits.logsumexp(dim=-1)
        return margins

    def get_out_to_loss_grad(self, model: Module, batch: Iterable[Tensor]) -> Tensor:
        """ Computes the (reweighting term Q in the paper)

        Args:
            model (torch.nn.Module)
                model
            batch (Iterable[Tensor]):
                input batch

        Returns:
            Tensor:
                out-to-loss (reweighting term) for the input batch
        """
        images, labels = batch
        logits = model(images)
        # here we are directly implementing the gradient instead of relying on autodiff to do
        # that for us
        ps = self.softmax(logits / self.loss_temperature)[ch.arange(logits.size(0)), labels]
        return (1 - ps).clone().detach().unsqueeze(-1)


class CLIPModelOutput(AbstractModelOutput):
    """ Margin for multimodal contrastive learning (CLIP). See Section 5.1 of
    `our paper <https://arxiv.org/abs/2303.14186>`_ for more details.

    Raises:
        AssertionError: this model output function requires using additional
        CLIP embeddings, which are computed using the :func:`get_embeddings`
        method. This method should be invoked before featurizing.
    """
    num_computed_embeddings = 0
    sim_batch_size = 0
    image_embeddings = None
    text_embeddings = None

    def __init__(self, temperature: float = None, simulated_batch_size: int = 300) -> None:
        """

        Args:
            temperature (float, optional):
                Temperature to use inside the softmax for the out-to-loss
                function. If None, CLIP's :code:`logit_scale` is used.  Defaults
                to None
            simulated_batch_size (int, optional):
                Size of the "simulated" batch size for the model output
                function. See Section 5.1 of the TRAK paper for more details.
                Defaults to 300.

        """
        super().__init__()
        self.softmax = ch.nn.Softmax(-1)
        self.temperature = temperature

        self.sim_batch_size = simulated_batch_size
        CLIPModelOutput.sim_batch_size = simulated_batch_size

    @staticmethod
    def get_embeddings(model,
                       loader,
                       batch_size: int,
                       size: int = 50_000,
                       embedding_dim: int = 1024,
                       preprocess_fn_img=None,
                       preprocess_fn_txt=None) -> None:
        """ Computes (image and text) embeddings and saves them in the class
        attributes :code:`image_embeddings` and :code:`text_embeddings`.

        Args:
            model (torch.nn.Module):
                model
            loader ():
                data loader
            batch_size (int):
                input batch size
            size (int, optional):
                Maximum number of embeddings to compute. Defaults to 50_000.
            embedding_dim (int, optional):
                Dimension of CLIP embedding. Defaults to 1024.
            preprocess_fn_img (func, optional):
                Transforms to apply to images from the loader before forward
                pass. Defaults to None.
            preprocess_fn_txt (func, optional):
                Transforms to apply to images from the loader before forward
                pass. Defaults to None.

        """
        img_embs, txt_embs = ch.zeros(size, embedding_dim).cuda(),\
            ch.zeros(size, embedding_dim).cuda()

        cutoff = batch_size
        with ch.no_grad():
            for ind, (images, text) in enumerate(loader):
                if preprocess_fn_img is not None:
                    images = preprocess_fn_img(images)
                if preprocess_fn_txt is not None:
                    text = preprocess_fn_txt(text)
                st, ed = ind * batch_size, min((ind + 1) * batch_size, size)
                if ed == size:
                    cutoff = size - ind * batch_size
                image_embeddings, text_embeddings, _ = model(images, text)
                img_embs[st: ed] = image_embeddings[:cutoff].clone().detach()
                txt_embs[st: ed] = text_embeddings[:cutoff].clone().detach()
                if (ind + 1) * batch_size >= size:
                    break

        CLIPModelOutput.image_embeddings = img_embs
        CLIPModelOutput.text_embeddings = txt_embs
        CLIPModelOutput.num_computed_embeddings = size

    @staticmethod
    def get_output(func_model,
                   weights: Iterable[Tensor],
                   buffers: Iterable[Tensor],
                   image: Tensor,
                   label: Tensor) -> Tensor:
        """ For a given input :math:`z=(x, y)` and model parameters
        :math:`\\theta`, let :math:`\\phi(x, \\theta)` be the CLIP image
        embedding and :math:`\\psi(y, \\theta)` be the CLIP text embedding.
        Last, let :math:`B` be a (simulated) batch. This method implements the
        model output function

        .. math::

            -\\log(\\frac{\\phi(x)\\cdot \\psi(y)}{\\sum_{(x', y')\\in B}
            \\phi(x)\\cdot \\psi(y')})
            -\\log(\\frac{\\phi(x)\\cdot \\psi(y)}{\\sum_{(x', y')\\in B}
            \\phi(x')\\cdot \\psi(y)})

        It uses functorch's functional models to make the per-sample gradient
        computations faster.  For more details on what func_model, weights &
        buffers are, and how to use them, please refer to
        https://pytorch.org/functorch/stable/ and
        https://pytorch.org/functorch/stable/notebooks/per_sample_grads.html.

        Note: this method slightly breaks abstraction since it has a different
        signature from the abstract :code:`get_output`. It's only used
        internally by :func:`GradientComputer` so it shouldn't be a big problem.
        If you're reading this and feel strongly about it, feel free to make a
        PR with a fix.

        Args:
            func_model (func):
                functorch functional model
            weights (Iterable[Tensor]):
                functorch model weights
            buffers (Iterable[Tensor]):
                functorch model buffers
            image (Tensor):
                input image, should not have batch dimension
            label (Tensor):
                input label, should not have batch dimension

        Returns:
            Tensor:
                model output for the given image-label pair :math:`z` and
                weights & buffers :math:`\\theta`.
        """
        all_im_embs = CLIPModelOutput.image_embeddings
        all_txt_embs = CLIPModelOutput.text_embeddings
        N = CLIPModelOutput.num_computed_embeddings
        sim_bs = CLIPModelOutput.sim_batch_size

        if all_im_embs is None:
            raise AssertionError('Run traker.modelout_fn.get_embeddings first before featurizing!')

        image_embeddings, text_embeddings, _ = func_model(weights,
                                                          buffers,
                                                          image.unsqueeze(0),
                                                          label.unsqueeze(0))

        ii = ch.multinomial(input=ch.arange(N).float(),
                            num_samples=sim_bs,
                            replacement=False)

        result = -ch.logsumexp(-image_embeddings @ (text_embeddings - all_txt_embs[ii]).T, dim=1) +\
                 -ch.logsumexp(-text_embeddings @ (image_embeddings - all_im_embs[ii]).T, dim=1)
        return result.sum()  # shape of result should be [1]

    def get_out_to_loss_grad(self, func_model, weights, buffers, batch: Iterable[Tensor]) -> Tensor:
        """ Computes the (reweighting term Q in the paper)

        Args:
            func_model (func):
                functorch functional model
            weights (Iterable[Tensor]):
                functorch model weights
            buffers (Iterable[Tensor]):
                functorch model buffers
            batch (Iterable[Tensor]):
                input batch

        Returns:
            Tensor:
                out-to-loss (reweighting term) for the input batch

        """
        image_embeddings, text_embeddings, temp = func_model(weights, buffers, *batch)
        if self.temperature is None:
            self.temperature = temp
        res = self.temperature * image_embeddings @ text_embeddings.T
        ps = (self.softmax(res) + self.softmax(res.T)).diag() / 2.
        return (1 - ps).clone().detach()


class TextClassificationModelOutput(AbstractModelOutput):
    """ Margin for text classification models. This assumes that the model takes
    in input_ids, token_type_ids, and attention_mask.

    .. math::

        \\text{logit}[\\text{correct}] - \\log\\left(\\sum_{i \\neq
        \\text{correct}} \\exp(\\text{logit}[i])\\right)

    """

    def __init__(self, temperature=1.) -> None:
        super().__init__()
        self.softmax = ch.nn.Softmax(-1)
        self.loss_temperature = temperature

    @staticmethod
    def get_output(func_model,
                   weights: Iterable[Tensor],
                   buffers: Iterable[Tensor],
                   input_id: Tensor,
                   token_type_id: Tensor,
                   attention_mask: Tensor,
                   label: Tensor,
                   ) -> Tensor:
        logits = func_model(weights, buffers, input_id.unsqueeze(0),
                            token_type_id.unsqueeze(0),
                            attention_mask.unsqueeze(0))
        bindex = ch.arange(logits.shape[0]).to(logits.device, non_blocking=False)
        logits_correct = logits[bindex, label.unsqueeze(0)]

        cloned_logits = logits.clone()
        cloned_logits[bindex, label.unsqueeze(0)] = ch.tensor(-ch.inf).to(logits.device)

        margins = logits_correct - cloned_logits.logsumexp(dim=-1)
        return margins.sum()

    def forward(self, model: Module, batch: Iterable[Tensor]) -> Tensor:
        input_ids, token_type_ids, attention_mask, _ = batch
        return model(input_ids=input_ids,
                     token_type_ids=token_type_ids,
                     attention_mask=attention_mask)

    def get_out_to_loss_grad(self, func_model, weights, buffers, batch: Iterable[Tensor]) -> Tensor:
        input_ids, token_type_ids, attention_mask, labels = batch
        logits = func_model(weights, buffers, input_ids, token_type_ids, attention_mask)
        ps = self.softmax(logits / self.loss_temperature)[ch.arange(logits.size(0)), labels]
        return (1 - ps).clone().detach().unsqueeze(-1)


TASK_TO_MODELOUT = {
    ('image_classification', True): ImageClassificationModelOutput,
    ('image_classification', False): IterImageClassificationModelOutput,
    ('text_classification', True): TextClassificationModelOutput,
    ('clip', True): CLIPModelOutput,
}



#!/usr/bin/env python
from setuptools import setup

setup(name="traker",
      version="0.1.3",
      description="TRAK: Attributing Model Behavior at Scale",
      long_description="Check https://trak.csail.mit.edu/ to learn more about TRAK",
      author="MadryLab",
      author_email='trak@mit.edu',
      license_files=('LICENSE.txt', ),
      packages=['trak'],
      install_requires=[
       "torch>=1.13",
       "numpy",
       "tqdm",
       ],
      extras_require={
          'tests':
              ["assertpy",
               "torchvision",
               "open_clip_torch",
               "wget",
               "scipy",
               ],
          'fast':
              ["fast_jl"
              ]},
      include_package_data=True,
      )



from abc import ABC, abstractmethod
from typing import Union
from enum import Enum
from torch import Tensor
import math
import torch
ch = torch


class ProjectionType(str, Enum):
    normal: str = 'normal'
    rademacher: str = 'rademacher'


class AbstractProjector(ABC):
    """ Implementations of the Projector class must implement the
    :meth:`AbstractProjector.project` method, which takes in model gradients and
    returns
    """
    @abstractmethod
    def __init__(self,
                 grad_dim: int,
                 proj_dim: int,
                 seed: int,
                 proj_type: Union[str, ProjectionType],
                 device: Union[str, torch.device]) -> None:
        """ Initializes hyperparameters for the projection.

        Args:
            grad_dim (int):
                number of parameters in the model (dimension of the gradient
                vectors)
            proj_dim (int):
                dimension after the projection
            seed (int):
                random seed for the generation of the sketching (projection)
                matrix
            proj_type (Union[str, ProjectionType]):
                the random projection (JL transform) guearantees that distances
                will be approximately preserved for a variety of choices of the
                random matrix (see e.g. https://arxiv.org/abs/1411.2404). Here,
                we provide an implementation for matrices with iid Gaussian
                entries and iid Rademacher entries.
            device (Union[str, torch.device]):
                CUDA device to use

        """
        self.grad_dim = grad_dim
        self.proj_dim = proj_dim
        self.seed = seed
        self.proj_type = proj_type
        self.device = device

    @abstractmethod
    def project(self, grads: Tensor, model_id: int) -> Tensor:
        """ Performs the random projection. Model ID is included
        so that we generate different projection matrices for every
        model ID.

        Args:
            grads (Tensor): a batch of gradients to be projected
            model_id (int): a unique ID for a checkpoint

        Returns:
            Tensor: the projected gradients
        """
        ...


class BasicSingleBlockProjector(AbstractProjector):
    """
    A bare-bones, inefficient implementation of the projection, which simply
    calls torch's matmul for the projection step.

    Note: for most model sizes (e.g. even for ResNet18), and small projection
    dimensions (e.g. anything > 100) this method will OOM on an A100.

    Unless you have a good reason to use this class (I cannot think of one, I
    added this only for testing purposes), use instead the CudaProjector or
    BasicProjector.
    """
    def __init__(self, grad_dim: int, proj_dim: int, seed: int, proj_type:
                 ProjectionType, device, dtype=ch.float16, model_id=0) -> None:
        super().__init__(grad_dim, proj_dim, seed, proj_type, device)

        self.model_id = model_id
        self.proj_type = proj_type
        self.generator = ch.Generator(device=self.device)
        self.generator = self.generator.manual_seed(self.seed + int(1e4) * self.model_id)
        self.dtype = dtype

        self.proj_matrix = ch.ones(self.grad_dim,
                                   self.proj_dim,
                                   dtype=self.dtype,
                                   device=self.device)

        self.generate_sketch_matrix()  # updates self.proj_matrix

    def generate_sketch_matrix(self):
        if self.proj_type == ProjectionType.normal or self.proj_type == 'normal':
            self.proj_matrix.normal_(generator=self.generator)
        elif self.proj_type == ProjectionType.rademacher or self.proj_type == 'rademacher':
            self.proj_matrix.bernoulli_(p=0.5, generator=self.generator)
            # going from Bernoulli {0, 1} to Rademacher {-1, 1}
            self.proj_matrix *= 2.
            self.proj_matrix -= 1.
        else:
            raise KeyError(f'Projection type {self.proj_type} not recognized.')

    def project(self, grads: Tensor, model_id: int) -> Tensor:
        if model_id != self.model_id:
            self.model_id = model_id
            self.generator = self.generator.manual_seed(self.seed + int(1e4) * self.model_id)
            self.generate_sketch_matrix()  # updates self.proj_matrix

        return grads @ self.proj_matrix


class BasicProjector(AbstractProjector):
    """
    A simple block-wise implementation of the projection. The projection matrix
    is generated on-device in blocks. The accumulated result across blocks is
    returned.

    Note: This class will be significantly slower and have a larger memory
    footprint than the CudaProjector. It is recommended that you use this method
    only if the CudaProjector is not available to you -- e.g. if you don't have
    a CUDA-enabled device with compute capability >=7.0 (see
    https://developer.nvidia.com/cuda-gpus).
    """
    def __init__(self, grad_dim: int, proj_dim: int, seed: int, proj_type:
                 ProjectionType, device, block_size: int = 200, dtype=ch.float32,
                 model_id=0) -> None:
        super().__init__(grad_dim, proj_dim, seed, proj_type, device)

        self.block_size = min(self.proj_dim, block_size)
        self.num_blocks = math.ceil(self.proj_dim / self.block_size)
        self.dtype = dtype
        self.proj_type = proj_type
        self.model_id = model_id

        self.proj_matrix = ch.empty(self.grad_dim, self.block_size,
                                    dtype=self.dtype,
                                    device=self.device)

        self.generator = ch.Generator(device=self.device)

        self.get_generator_states()
        self.generate_sketch_matrix(self.generator_states[0])

    def get_generator_states(self):
        self.generator_states = []
        self.seeds = []
        self.jl_size = self.proj_matrix.numel()

        for i in range(self.num_blocks):
            s = self.seed + int(1e3) * i + int(1e5) * self.model_id
            self.seeds.append(s)
            self.generator = self.generator.manual_seed(s)
            self.generator_states.append(self.generator.get_state())

    def generate_sketch_matrix(self, generator_state):
        self.generator.set_state(generator_state)
        if self.proj_type == ProjectionType.normal or self.proj_type == 'normal':
            self.proj_matrix.normal_(generator=self.generator)
        elif self.proj_type == ProjectionType.rademacher or self.proj_type == 'rademacher':
            self.proj_matrix.bernoulli_(p=0.5, generator=self.generator)
            self.proj_matrix *= 2.
            self.proj_matrix -= 1.
        else:
            raise KeyError(f'Projection type {self.proj_type} not recognized.')

    def project(self, grads: Tensor, model_id: int) -> Tensor:
        sketch = ch.zeros(size=(grads.size(0), self.proj_dim),
                          dtype=self.dtype, device=self.device)

        if model_id != self.model_id:
            self.model_id = model_id
            self.get_generator_states()  # regenerate random seeds for new model_id
            if self.num_blocks == 1:
                self.generate_sketch_matrix(self.generator_states[0])

        if self.num_blocks == 1:
            ch.matmul(grads.data, self.proj_matrix, out=sketch)
        else:
            for ind in range(self.num_blocks):
                self.generate_sketch_matrix(self.generator_states[ind])

                st = ind * self.block_size
                ed = min((ind + 1) * self.block_size, self.proj_dim)
                sketch[:, st:ed] = grads.type(self.dtype) @ self.proj_matrix[:, :(ed - st)]
        return sketch.type(grads.dtype)


class CudaProjector(AbstractProjector):
    """
    A performant implementation of the projection for CUDA with compute
    capability >= 7.0.
    """
    def __init__(self, grad_dim: int, proj_dim: int, seed: int, proj_type:
                 ProjectionType, device, max_batch_size: int = 32, *args, **kwargs) -> None:
        """

        Args:
            grad_dim (int):
                Number of parameters
            proj_dim (int):
                Dimension we project *to* during the projection step
            seed (int):
                Random seed
            proj_type (ProjectionType):
                Type of randomness to use for projection matrix (rademacher or normal)
            device:
                CUDA device
            max_batch_size (int):
                Explicitly constraints the batch size the CudaProjector is going
                to use for projection. Set this if you get a 'The batch size of
                the CudaProjector is too large for your GPU' error. Must be
                either 8, 16, or 32.

        Raises:
            ValueError:
                When attempting to use this on a non-CUDA device
            ModuleNotFoundError:
                When fast_jl is not installed

        """
        super().__init__(grad_dim, proj_dim, seed, proj_type, device)
        self.max_batch_size = max_batch_size

        if isinstance(device, str):
            device = ch.device(device)

        if device.type != 'cuda':
            err = "CudaProjector only works on a CUDA device; Either switch to a CUDA device, or use the BasicProjector"
            raise ValueError(err)

        self.num_sms = ch.cuda.get_device_properties(device.index).multi_processor_count

        try:
            import fast_jl
            # test run to catch at init time if projection goes through
            fast_jl.project_rademacher_8(ch.zeros(8, 1_000, device='cuda'), 512, 0, self.num_sms)
        except ImportError:
            err = "You should make sure to install the CUDA projector for traker (called fast_jl).\
                  See the installation FAQs for more details."
            raise ModuleNotFoundError(err)

    def project(self, grads: Tensor, model_id: int) -> Tensor:
        batch_size = grads.shape[0]

        effective_batch_size = 32
        if batch_size <= 8:
            effective_batch_size = 8
        elif batch_size <= 16:
            effective_batch_size = 16

        effective_batch_size = min(self.max_batch_size, effective_batch_size)

        function_name = f"project_{self.proj_type.value}_{effective_batch_size}"
        import fast_jl
        fn = getattr(fast_jl, function_name)

        try:
            result = fn(grads, self.proj_dim, self.seed + int(1e4) * model_id, self.num_sms)
        except RuntimeError as e:
            if str(e) == 'CUDA error: too many resources requested for launch\nCUDA kernel errors might be asynchronously reported at some other API call, so the stacktrace below might be incorrect.\nFor debugging consider passing CUDA_LAUNCH_BLOCKING=1.\nCompile with `TORCH_USE_CUDA_DSA` to enable device-side assertions.\n':
                # provide a more helpful error message
                raise RuntimeError(f'The batch size of the CudaProjector is too large for your GPU. Reduce it by using the max_batch_size argument of the CudaProjector.\nOriginal error: {str(e)}')
            else:
                raise e

        return result



from abc import ABC, abstractmethod
from torch import Tensor
import torch as ch


class AbstractScoreComputer(ABC):
    """
    The :code:`ScoreComputer` class
    Implementations of the ScoreComputer class must implement three methods:
    - :code:`get_xtx`
    - :code:`get_x_xtx_inv`
    - :code:`get_scores`
    """
    @abstractmethod
    def __init__(self, device) -> None:
        self.device = device

    @abstractmethod
    def get_xtx(self, grads: Tensor) -> Tensor:
        ...

    @abstractmethod
    def get_x_xtx_inv(self, grads: Tensor, xtx: Tensor) -> Tensor:
        ...

    @abstractmethod
    def get_scores(self, features: Tensor, target_grads: Tensor) -> Tensor:
        ...


class BasicSingleBlockScoreComputer(AbstractScoreComputer):
    """ A bare-bones implementation of :code:`ScoreComputer` that will likely
    OOM for almost all applications. Here for testing purposes only. Unless you
    have a good reason not to, you should use :func:`BasicScoreComputer`
    instead.
    """
    def __init__(self, device) -> None:
        super().__init__(device)

    def get_xtx(self, grads: Tensor) -> Tensor:
        return grads.T @ grads

    def get_x_xtx_inv(self, grads: Tensor, xtx: Tensor) -> Tensor:
        return grads @ ch.linalg.inv(xtx)

    def get_scores(self, features: Tensor, target_grads: Tensor) -> Tensor:
        return features @ target_grads.T


class BasicScoreComputer(AbstractScoreComputer):
    """ An implementation of :code:`ScoreComputer` that computes matmuls in a
    block-wise manner.
    """
    def __init__(self, device, CUDA_MAX_DIM_SIZE: int = 100_000) -> None:
        """
        Args:
            device (Union[str, torch.device]): torch device to do matmuls on
            CUDA_MAX_DIM_SIZE (int, optional): Size of block for block-wise
            matmuls. Defaults to 100_000.
        """
        super().__init__(device)
        self.CUDA_MAX_DIM_SIZE = CUDA_MAX_DIM_SIZE

    def get_xtx(self, grads: Tensor) -> Tensor:
        self.proj_dim = grads.shape[1]
        result = ch.zeros(self.proj_dim, self.proj_dim).to(self.device)
        blocks = ch.split(grads, split_size_or_sections=self.CUDA_MAX_DIM_SIZE, dim=0)

        for block in blocks:
            result += block.T.to(self.device) @ block.to(self.device)

        return result

    def get_x_xtx_inv(self, grads: Tensor, xtx: Tensor) -> Tensor:
        blocks = ch.split(grads, split_size_or_sections=self.CUDA_MAX_DIM_SIZE, dim=0)
        xtx_inv = ch.linalg.inv(xtx)

        result = ch.empty(grads.shape[0], xtx_inv.shape[1], device=self.device)
        for i, block in enumerate(blocks):
            start = i * self.CUDA_MAX_DIM_SIZE
            end = min(grads.shape[0], (i + 1) * self.CUDA_MAX_DIM_SIZE)
            result[start: end] = (block.to(self.device) @ xtx_inv)
        return result

    def get_scores(self, features: Tensor, target_grads: Tensor) -> Tensor:
        train_dim = features.shape[0]
        target_dim = target_grads.shape[0]

        if target_dim < self.CUDA_MAX_DIM_SIZE:
            return features @ target_grads.T

        result = ch.empty(train_dim, target_dim).to(self.device)
        blocks = ch.split(target_grads, split_size_or_sections=self.CUDA_MAX_DIM_SIZE, dim=0)

        for i, block in enumerate(blocks):
            start = i * self.CUDA_MAX_DIM_SIZE
            end = min(target_grads.shape[0], (i + 1) * self.CUDA_MAX_DIM_SIZE)
            result[:, start: end] = features @ block.T

        return result
